@@ -89,7 +89,7 @@
 #endif // USE_OBSOLETE_LINKER
 
 /* use this attribute on a variable to move it to the read_mostly section */
-#if !defined(__DragonFly__)
+#if !defined(__read_mostly)
 #define __read_mostly           HA_SECTION("read_mostly")
 #endif
 
@@ -102,7 +102,7 @@
 #ifdef DEBUG_USE_ABORT
 #define my_unreachable() abort()
 #else
-#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 5)
+#if defined(__GNUC__) && (__GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 5))
 #define my_unreachable() __builtin_unreachable()
 #else
 #define my_unreachable()
