@@ -130,14 +130,12 @@ int ssl_sock_load_sctl_from_file(const char *sctl_path, char *buf, struct cert_k
 
 	sctl = calloc(1, sizeof(*sctl));
 	if (!chunk_dup(sctl, src)) {
-		free(sctl);
-		sctl = NULL;
+		ha_free(&sctl);
 		goto end;
 	}
 	/* no error, fill ckch with new context, old context must be free */
 	if (ckch->sctl) {
-		free(ckch->sctl->area);
-		ckch->sctl->area = NULL;
+		ha_free(&ckch->sctl->area);
 		free(ckch->sctl);
 	}
 	ckch->sctl = sctl;
@@ -212,14 +210,12 @@ int ssl_sock_load_ocsp_response_from_file(const char *ocsp_path, char *buf, stru
 
 	ocsp_response = calloc(1, sizeof(*ocsp_response));
 	if (!chunk_dup(ocsp_response, src)) {
-		free(ocsp_response);
-		ocsp_response = NULL;
+		ha_free(&ocsp_response);
 		goto end;
 	}
 	/* no error, fill ckch with new context, old context must be free */
 	if (ckch->ocsp_response) {
-		free(ckch->ocsp_response->area);
-		ckch->ocsp_response->area = NULL;
+		ha_free(&ckch->ocsp_response->area);
 		free(ckch->ocsp_response);
 	}
 	ckch->ocsp_response = ocsp_response;
@@ -277,7 +273,7 @@ int ssl_sock_load_files_into_ckch(const char *path, struct cert_key_and_chain *c
 		/* look for the extension */
 		if ((ext = strrchr(fp->area, '.'))) {
 
-			if (!strcmp(ext, ".crt")) {
+			if (strcmp(ext, ".crt") == 0) {
 				*ext = '\0';
 				fp->data = strlen(fp->area);
 			}
@@ -562,17 +558,13 @@ int ssl_sock_load_pem_into_ckch(const char *path, char *buf, struct cert_key_and
 
 	/* once it loaded the PEM, it should remove everything else in the ckch */
 	if (ckch->ocsp_response) {
-		free(ckch->ocsp_response->area);
-		ckch->ocsp_response->area = NULL;
-		free(ckch->ocsp_response);
-		ckch->ocsp_response = NULL;
+		ha_free(&ckch->ocsp_response->area);
+		ha_free(&ckch->ocsp_response);
 	}
 
 	if (ckch->sctl) {
-		free(ckch->sctl->area);
-		ckch->sctl->area = NULL;
-		free(ckch->sctl);
-		ckch->sctl = NULL;
+		ha_free(&ckch->sctl->area);
+		ha_free(&ckch->sctl);
 	}
 
 	if (ckch->ocsp_issuer) {
@@ -632,17 +624,13 @@ void ssl_sock_free_cert_key_and_chain_contents(struct cert_key_and_chain *ckch)
 	ckch->dh = NULL;
 
 	if (ckch->sctl) {
-		free(ckch->sctl->area);
-		ckch->sctl->area = NULL;
-		free(ckch->sctl);
-		ckch->sctl = NULL;
+		ha_free(&ckch->sctl->area);
+		ha_free(&ckch->sctl);
 	}
 
 	if (ckch->ocsp_response) {
-		free(ckch->ocsp_response->area);
-		ckch->ocsp_response->area = NULL;
-		free(ckch->ocsp_response);
-		ckch->ocsp_response = NULL;
+		ha_free(&ckch->ocsp_response->area);
+		ha_free(&ckch->ocsp_response);
 	}
 
 	if (ckch->ocsp_issuer)
@@ -689,8 +677,7 @@ struct cert_key_and_chain *ssl_sock_copy_cert_key_and_chain(struct cert_key_and_
 
 		sctl = calloc(1, sizeof(*sctl));
 		if (!chunk_dup(sctl, src->sctl)) {
-			free(sctl);
-			sctl = NULL;
+			ha_free(&sctl);
 			goto error;
 		}
 		dst->sctl = sctl;
@@ -701,8 +688,7 @@ struct cert_key_and_chain *ssl_sock_copy_cert_key_and_chain(struct cert_key_and_
 
 		ocsp_response = calloc(1, sizeof(*ocsp_response));
 		if (!chunk_dup(ocsp_response, src->ocsp_response)) {
-			free(ocsp_response);
-			ocsp_response = NULL;
+			ha_free(&ocsp_response);
 			goto error;
 		}
 		dst->ocsp_response = ocsp_response;
@@ -788,8 +774,7 @@ void ckch_store_free(struct ckch_store *store)
 
 	ssl_sock_free_cert_key_and_chain_contents(store->ckch);
 
-	free(store->ckch);
-	store->ckch = NULL;
+	ha_free(&store->ckch);
 
 	list_for_each_entry_safe(inst, inst_s, &store->ckch_inst, by_ckchs) {
 		ckch_inst_free(inst);
@@ -909,14 +894,14 @@ void ckch_inst_free(struct ckch_inst *inst)
 
 	list_for_each_entry_safe(sni, sni_s, &inst->sni_ctx, by_ckch_inst) {
 		SSL_CTX_free(sni->ctx);
-		LIST_DEL(&sni->by_ckch_inst);
+		LIST_DELETE(&sni->by_ckch_inst);
 		ebmb_delete(&sni->name);
 		free(sni);
 	}
 	SSL_CTX_free(inst->ctx);
 	inst->ctx = NULL;
-	LIST_DEL(&inst->by_ckchs);
-	LIST_DEL(&inst->by_crtlist_entry);
+	LIST_DELETE(&inst->by_ckchs);
+	LIST_DELETE(&inst->by_crtlist_entry);
 	free(inst);
 }
 
@@ -1228,7 +1213,7 @@ static int cli_parse_show_cert(char **args, char *payload, struct appctx *appctx
 
 			ckchs = ckchs_transaction.new_ckchs;
 
-			if (strcmp(args[3] + 1, ckchs->path))
+			if (strcmp(args[3] + 1, ckchs->path) != 0)
 				goto error;
 
 		} else {
@@ -1274,6 +1259,7 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 	int y = 0;
 	char *err = NULL;
 	int errcode = 0;
+	int retval = 0;
 	struct ckch_store *old_ckchs, *new_ckchs = NULL;
 	struct ckch_inst *ckchi, *ckchis;
 	struct buffer *trash = alloc_trash_chunk();
@@ -1336,7 +1322,10 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 						fcount = ckchi->crtlist_entry->fcount;
 					}
 
-					errcode |= ckch_inst_new_load_store(new_ckchs->path, new_ckchs, ckchi->bind_conf, ckchi->ssl_conf, sni_filter, fcount, &new_inst, &err);
+					if (ckchi->is_server_instance)
+						errcode |= ckch_inst_new_load_srv_store(new_ckchs->path, new_ckchs, &new_inst, &err);
+					else
+						errcode |= ckch_inst_new_load_store(new_ckchs->path, new_ckchs, ckchi->bind_conf, ckchi->ssl_conf, sni_filter, fcount, &new_inst, &err);
 
 					if (errcode & ERR_CODE)
 						goto error;
@@ -1345,6 +1334,15 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 					if (ckchi->is_default)
 						new_inst->is_default = 1;
 
+					new_inst->is_server_instance = ckchi->is_server_instance;
+					new_inst->server = ckchi->server;
+					/* Create a new SSL_CTX and link it to the new instance. */
+					if (new_inst->is_server_instance) {
+						retval = ssl_sock_prepare_srv_ssl_ctx(ckchi->server, new_inst->ctx);
+						if (retval)
+							goto error;
+					}
+
 					/* create the link to the crtlist_entry */
 					new_inst->crtlist_entry = ckchi->crtlist_entry;
 
@@ -1362,7 +1360,7 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 					/* display one dot per new instance */
 					chunk_appendf(trash, ".");
 					/* link the new ckch_inst to the duplicate */
-					LIST_ADDQ(&new_ckchs->ckch_inst, &new_inst->by_ckchs);
+					LIST_APPEND(&new_ckchs->ckch_inst, &new_inst->by_ckchs);
 					y++;
 				}
 				appctx->st2 = SETCERT_ST_INSERT;
@@ -1388,23 +1386,49 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 				/* insert the new ckch_insts in the crtlist_entry */
 				list_for_each_entry(ckchi, &new_ckchs->ckch_inst, by_ckchs) {
 					if (ckchi->crtlist_entry)
-						LIST_ADD(&ckchi->crtlist_entry->ckch_inst, &ckchi->by_crtlist_entry);
+						LIST_INSERT(&ckchi->crtlist_entry->ckch_inst, &ckchi->by_crtlist_entry);
 				}
 
 				/* First, we insert every new SNIs in the trees, also replace the default_ctx */
 				list_for_each_entry_safe(ckchi, ckchis, &new_ckchs->ckch_inst, by_ckchs) {
-					HA_RWLOCK_WRLOCK(SNI_LOCK, &ckchi->bind_conf->sni_lock);
-					ssl_sock_load_cert_sni(ckchi, ckchi->bind_conf);
-					HA_RWLOCK_WRUNLOCK(SNI_LOCK, &ckchi->bind_conf->sni_lock);
+					/* The bind_conf will be null on server ckch_instances. */
+					if (ckchi->is_server_instance) {
+						int i;
+						/* a lock is needed here since we have to free the SSL cache */
+						HA_RWLOCK_WRLOCK(SSL_SERVER_LOCK, &ckchi->server->ssl_ctx.lock);
+						/* free the server current SSL_CTX */
+						SSL_CTX_free(ckchi->server->ssl_ctx.ctx);
+						/* Actual ssl context update */
+						SSL_CTX_up_ref(ckchi->ctx);
+						ckchi->server->ssl_ctx.ctx = ckchi->ctx;
+						ckchi->server->ssl_ctx.inst = ckchi;
+
+						/* flush the session cache of the server */
+						for (i = 0; i < global.nbthread; i++)
+							ha_free(&ckchi->server->ssl_ctx.reused_sess[i].ptr);
+
+						HA_RWLOCK_WRUNLOCK(SSL_SERVER_LOCK, &ckchi->server->ssl_ctx.lock);
+
+					} else {
+						HA_RWLOCK_WRLOCK(SNI_LOCK, &ckchi->bind_conf->sni_lock);
+						ssl_sock_load_cert_sni(ckchi, ckchi->bind_conf);
+						HA_RWLOCK_WRUNLOCK(SNI_LOCK, &ckchi->bind_conf->sni_lock);
+					}
 				}
 
 				/* delete the old sni_ctx, the old ckch_insts and the ckch_store */
 				list_for_each_entry_safe(ckchi, ckchis, &old_ckchs->ckch_inst, by_ckchs) {
-					struct bind_conf __maybe_unused *bind_conf = ckchi->bind_conf;
 
-					HA_RWLOCK_WRLOCK(SNI_LOCK, &bind_conf->sni_lock);
-					ckch_inst_free(ckchi);
-					HA_RWLOCK_WRUNLOCK(SNI_LOCK, &bind_conf->sni_lock);
+					if (ckchi->is_server_instance) {
+						/* no lock for servers */
+						ckch_inst_free(ckchi);
+					} else {
+						struct bind_conf __maybe_unused *bind_conf = ckchi->bind_conf;
+
+						HA_RWLOCK_WRLOCK(SNI_LOCK, &bind_conf->sni_lock);
+						ckch_inst_free(ckchi);
+						HA_RWLOCK_WRUNLOCK(SNI_LOCK, &bind_conf->sni_lock);
+					}
 				}
 
 				/* Replace the old ckchs by the new one */
@@ -1414,8 +1438,7 @@ static int cli_io_handler_commit_cert(struct appctx *appctx)
 				/* fallthrough */
 			case SETCERT_ST_FIN:
 				/* we achieved the transaction, we can set everything to NULL */
-				free(ckchs_transaction.path);
-				ckchs_transaction.path = NULL;
+				ha_free(&ckchs_transaction.path);
 				ckchs_transaction.new_ckchs = NULL;
 				ckchs_transaction.old_ckchs = NULL;
 				goto end;
@@ -1537,8 +1560,11 @@ static int cli_parse_set_cert(char **args, char *payload, struct appctx *appctx,
 	if (HA_SPIN_TRYLOCK(CKCH_LOCK, &ckch_lock))
 		return cli_err(appctx, "Can't update the certificate!\nOperations on certificates are currently locked!\n");
 
-	if ((buf = alloc_trash_chunk()) == NULL)
-		return cli_err(appctx, "Can't allocate memory\n");
+	if ((buf = alloc_trash_chunk()) == NULL) {
+		memprintf(&err, "%sCan't allocate memory\n", err ? err : "");
+		errcode |= ERR_ALERT | ERR_FATAL;
+		goto end;
+	}
 
 	if (!chunk_strcpy(buf, args[3])) {
 		memprintf(&err, "%sCan't allocate memory\n", err ? err : "");
@@ -1549,7 +1575,7 @@ static int cli_parse_set_cert(char **args, char *payload, struct appctx *appctx,
 	/* check which type of file we want to update */
 	for (i = 0; cert_exts[i].type < CERT_TYPE_MAX; i++) {
 		end = strrchr(buf->area, '.');
-		if (end && *cert_exts[i].ext && (!strcmp(end + 1, cert_exts[i].ext))) {
+		if (end && *cert_exts[i].ext && (strcmp(end + 1, cert_exts[i].ext) == 0)) {
 			*end = '\0';
 			buf->data = strlen(buf->area);
 			type = cert_exts[i].type;
@@ -1675,8 +1701,7 @@ end:
 
 		appctx->ctx.ssl.old_ckchs = NULL;
 
-		free(appctx->ctx.ssl.path);
-		appctx->ctx.ssl.path = NULL;
+		ha_free(&appctx->ctx.ssl.path);
 
 		HA_SPIN_UNLOCK(CKCH_LOCK, &ckch_lock);
 		return cli_dynerr(appctx, memprintf(&err, "%sCan't update %s!\n", err ? err : "", args[3]));
@@ -1718,8 +1743,7 @@ static int cli_parse_abort_cert(char **args, char *payload, struct appctx *appct
 	ckch_store_free(ckchs_transaction.new_ckchs);
 	ckchs_transaction.new_ckchs = NULL;
 	ckchs_transaction.old_ckchs = NULL;
-	free(ckchs_transaction.path);
-	ckchs_transaction.path = NULL;
+	ha_free(&ckchs_transaction.path);
 
 	HA_SPIN_UNLOCK(CKCH_LOCK, &ckch_lock);
 
@@ -1835,12 +1859,12 @@ void ckch_deinit()
 
 /* register cli keywords */
 static struct cli_kw_list cli_kws = {{ },{
-	{ { "new", "ssl", "cert", NULL }, "new ssl cert <certfile> : create a new certificate file to be used in a crt-list or a directory", cli_parse_new_cert, NULL, NULL },
-	{ { "set", "ssl", "cert", NULL }, "set ssl cert <certfile> <payload> : replace a certificate file", cli_parse_set_cert, NULL, NULL },
-	{ { "commit", "ssl", "cert", NULL }, "commit ssl cert <certfile> : commit a certificate file", cli_parse_commit_cert, cli_io_handler_commit_cert, cli_release_commit_cert },
-	{ { "abort", "ssl", "cert", NULL }, "abort ssl cert <certfile> : abort a transaction for a certificate file", cli_parse_abort_cert, NULL, NULL },
-	{ { "del", "ssl", "cert", NULL }, "del ssl cert <certfile> : delete an unused certificate file", cli_parse_del_cert, NULL, NULL },
-	{ { "show", "ssl", "cert", NULL }, "show ssl cert [<certfile>] : display the SSL certificates used in memory, or the details of a <certfile>", cli_parse_show_cert, cli_io_handler_show_cert, cli_release_show_cert },
+	{ { "new", "ssl", "cert", NULL },    "new ssl cert <certfile>                 : create a new certificate file to be used in a crt-list or a directory", cli_parse_new_cert, NULL, NULL },
+	{ { "set", "ssl", "cert", NULL },    "set ssl cert <certfile> <payload>       : replace a certificate file",                                            cli_parse_set_cert, NULL, NULL },
+	{ { "commit", "ssl", "cert", NULL }, "commit ssl cert <certfile>              : commit a certificate file",                                             cli_parse_commit_cert, cli_io_handler_commit_cert, cli_release_commit_cert },
+	{ { "abort", "ssl", "cert", NULL },  "abort ssl cert <certfile>               : abort a transaction for a certificate file",                            cli_parse_abort_cert, NULL, NULL },
+	{ { "del", "ssl", "cert", NULL },    "del ssl cert <certfile>                 : delete an unused certificate file",                                     cli_parse_del_cert, NULL, NULL },
+	{ { "show", "ssl", "cert", NULL },   "show ssl cert [<certfile>]              : display the SSL certificates used in memory, or the details of a file", cli_parse_show_cert, cli_io_handler_show_cert, cli_release_show_cert },
 	{ { NULL }, NULL, NULL, NULL }
 }};
 
