@@ -18,14 +18,12 @@
 #include <haproxy/api.h>
 #include <haproxy/cfgparse.h>
 #include <haproxy/fd.h>
-#include <haproxy/freq_ctr.h>
 #include <haproxy/list.h>
 #include <haproxy/pool.h>
-#include <haproxy/stream.h>
 #include <haproxy/task.h>
-#include <haproxy/time.h>
 #include <haproxy/tools.h>
 
+extern struct task *process_stream(struct task *t, void *context, unsigned int state);
 
 DECLARE_POOL(pool_head_task,    "task",    sizeof(struct task));
 DECLARE_POOL(pool_head_tasklet, "tasklet", sizeof(struct tasklet));
@@ -35,9 +33,7 @@ DECLARE_POOL(pool_head_tasklet, "tasklet", sizeof(struct tasklet));
  */
 DECLARE_POOL(pool_head_notification, "notification", sizeof(struct notification));
 
-unsigned int nb_tasks = 0;
 volatile unsigned long global_tasks_mask = 0; /* Mask of threads with tasks in the global runqueue */
-unsigned int nb_tasks_cur = 0;     /* copy of the tasks count */
 unsigned int niced_tasks = 0;      /* number of niced tasks in the run queue */
 
 THREAD_LOCAL struct task_per_thread *sched = &task_per_thread[0]; /* scheduler context for the current thread */
@@ -48,10 +44,10 @@ __decl_aligned_rwlock(wq_lock);   /* RW lock related to the wait queue */
 #ifdef USE_THREAD
 struct eb_root timers;      /* sorted timers tree, global, accessed under wq_lock */
 struct eb_root rqueue;      /* tree constituting the global run queue, accessed under rq_lock */
-unsigned int grq_total;     /* total number of entries in the global run queue, use grq_lock */
+unsigned int grq_total;     /* total number of entries in the global run queue, atomic */
+static unsigned int global_rqueue_ticks;  /* insertion count in the grq, use rq_lock */
 #endif
 
-static unsigned int rqueue_ticks;  /* insertion count */
 
 struct task_per_thread task_per_thread[MAX_THREADS];
 
@@ -94,10 +90,10 @@ void task_kill(struct task *t)
 			thr = my_ffsl(t->thread_mask) - 1;
 
 			/* Beware: tasks that have never run don't have their ->list empty yet! */
-			MT_LIST_ADDQ(&task_per_thread[thr].shared_tasklet_list,
+			MT_LIST_APPEND(&task_per_thread[thr].shared_tasklet_list,
 			             (struct mt_list *)&((struct tasklet *)t)->list);
-			_HA_ATOMIC_ADD(&task_per_thread[thr].rq_total, 1);
-			_HA_ATOMIC_ADD(&task_per_thread[thr].task_list_size, 1);
+			_HA_ATOMIC_INC(&task_per_thread[thr].rq_total);
+			_HA_ATOMIC_INC(&task_per_thread[thr].tasks_in_list);
 			if (sleeping_thread_mask & (1UL << thr)) {
 				_HA_ATOMIC_AND(&sleeping_thread_mask, ~(1UL << thr));
 				wake_thread(thr);
@@ -107,6 +103,48 @@ void task_kill(struct task *t)
 	}
 }
 
+/* Do not call this one, please use tasklet_wakeup_on() instead, as this one is
+ * the slow path of tasklet_wakeup_on() which performs some preliminary checks
+ * and sets TASK_IN_LIST before calling this one. A negative <thr> designates
+ * the current thread.
+ */
+void __tasklet_wakeup_on(struct tasklet *tl, int thr)
+{
+	if (likely(thr < 0)) {
+		/* this tasklet runs on the caller thread */
+		if (tl->state & TASK_HEAVY) {
+			LIST_APPEND(&sched->tasklets[TL_HEAVY], &tl->list);
+			sched->tl_class_mask |= 1 << TL_HEAVY;
+		}
+		else if (tl->state & TASK_SELF_WAKING) {
+			LIST_APPEND(&sched->tasklets[TL_BULK], &tl->list);
+			sched->tl_class_mask |= 1 << TL_BULK;
+		}
+		else if ((struct task *)tl == sched->current) {
+			_HA_ATOMIC_OR(&tl->state, TASK_SELF_WAKING);
+			LIST_APPEND(&sched->tasklets[TL_BULK], &tl->list);
+			sched->tl_class_mask |= 1 << TL_BULK;
+		}
+		else if (sched->current_queue < 0) {
+			LIST_APPEND(&sched->tasklets[TL_URGENT], &tl->list);
+			sched->tl_class_mask |= 1 << TL_URGENT;
+		}
+		else {
+			LIST_APPEND(&sched->tasklets[sched->current_queue], &tl->list);
+			sched->tl_class_mask |= 1 << sched->current_queue;
+		}
+		_HA_ATOMIC_INC(&sched->rq_total);
+	} else {
+		/* this tasklet runs on a specific thread. */
+		MT_LIST_APPEND(&task_per_thread[thr].shared_tasklet_list, (struct mt_list *)&tl->list);
+		_HA_ATOMIC_INC(&task_per_thread[thr].rq_total);
+		if (sleeping_thread_mask & (1UL << thr)) {
+			_HA_ATOMIC_AND(&sleeping_thread_mask, ~(1UL << thr));
+			wake_thread(thr);
+		}
+	}
+}
+
 /* Puts the task <t> in run queue at a position depending on t->nice. <t> is
  * returned. The nice value assigns boosts in 32th of the run queue size. A
  * nice value of -1024 sets the task to -tasks_run_queue*32, while a nice value
@@ -115,28 +153,31 @@ void task_kill(struct task *t)
  * The task must not already be in the run queue. If unsure, use the safer
  * task_wakeup() function.
  */
-void __task_wakeup(struct task *t, struct eb_root *root)
+void __task_wakeup(struct task *t)
 {
+	struct eb_root *root = &sched->rqueue;
+
 #ifdef USE_THREAD
-	if (root == &rqueue) {
+	if (t->thread_mask != tid_bit && global.nbthread != 1) {
+		root = &rqueue;
+
+		_HA_ATOMIC_INC(&grq_total);
 		HA_SPIN_LOCK(TASK_RQ_LOCK, &rq_lock);
-	}
 
-	if (root == &rqueue) {
 		global_tasks_mask |= t->thread_mask;
-		grq_total++;
+		t->rq.key = ++global_rqueue_ticks;
 		__ha_barrier_store();
 	} else
 #endif
 	{
-		_HA_ATOMIC_ADD(&sched->rq_total, 1);
+		_HA_ATOMIC_INC(&sched->rq_total);
+		t->rq.key = ++sched->rqueue_ticks;
 	}
-	t->rq.key = _HA_ATOMIC_ADD(&rqueue_ticks, 1);
 
 	if (likely(t->nice)) {
 		int offset;
 
-		_HA_ATOMIC_ADD(&niced_tasks, 1);
+		_HA_ATOMIC_INC(&niced_tasks);
 		offset = t->nice * (int)global.tune.runqueue_depth;
 		t->rq.key += offset;
 	}
@@ -145,27 +186,23 @@ void __task_wakeup(struct task *t, struct eb_root *root)
 		t->call_date = now_mono_time();
 
 	eb32sc_insert(root, &t->rq, t->thread_mask);
+
 #ifdef USE_THREAD
 	if (root == &rqueue) {
 		_HA_ATOMIC_OR(&t->state, TASK_GLOBAL);
 		HA_SPIN_UNLOCK(TASK_RQ_LOCK, &rq_lock);
-	} else
-#endif
-	{
-		int nb = ((void *)root - (void *)&task_per_thread[0].rqueue) / sizeof(task_per_thread[0]);
-		task_per_thread[nb].rqueue_size++;
-	}
-#ifdef USE_THREAD
-	/* If all threads that are supposed to handle this task are sleeping,
-	 * wake one.
-	 */
-	if ((((t->thread_mask & all_threads_mask) & sleeping_thread_mask) ==
-	     (t->thread_mask & all_threads_mask))) {
-		unsigned long m = (t->thread_mask & all_threads_mask) &~ tid_bit;
 
-		m = (m & (m - 1)) ^ m; // keep lowest bit set
-		_HA_ATOMIC_AND(&sleeping_thread_mask, ~m);
-		wake_thread(my_ffsl(m) - 1);
+		/* If all threads that are supposed to handle this task are sleeping,
+		 * wake one.
+		 */
+		if ((((t->thread_mask & all_threads_mask) & sleeping_thread_mask) ==
+		     (t->thread_mask & all_threads_mask))) {
+			unsigned long m = (t->thread_mask & all_threads_mask) &~ tid_bit;
+
+			m = (m & (m - 1)) ^ m; // keep lowest bit set
+			_HA_ATOMIC_AND(&sleeping_thread_mask, ~m);
+			wake_thread(my_ffsl(m) - 1);
+		}
 	}
 #endif
 	return;
@@ -322,29 +359,12 @@ void wake_expired_tasks()
 		}
 
 		task = eb32_entry(eb, struct task, wq);
-
-		/* Check for any competing run of the task (quite rare but may
-		 * involve a dangerous concurrent access on task->expire). In
-		 * order to protect against this, we'll take an exclusive access
-		 * on TASK_RUNNING before checking/touching task->expire. If the
-		 * task is already RUNNING on another thread, it will deal by
-		 * itself with the requeuing so we must not do anything and
-		 * simply quit the loop for now, because we cannot wait with the
-		 * WQ lock held as this would prevent the running thread from
-		 * requeuing the task. One annoying effect of holding RUNNING
-		 * here is that a concurrent task_wakeup() will refrain from
-		 * waking it up. This forces us to check for a wakeup after
-		 * releasing the flag.
-		 */
-		if (HA_ATOMIC_FETCH_OR(&task->state, TASK_RUNNING) & TASK_RUNNING)
-			break;
-
 		if (tick_is_expired(task->expire, now_ms)) {
 			/* expired task, wake it up */
 			HA_RWLOCK_SKTOWR(TASK_WQ_LOCK, &wq_lock);
 			__task_unlink_wq(task);
 			HA_RWLOCK_WRTOSK(TASK_WQ_LOCK, &wq_lock);
-			task_drop_running(task, TASK_WOKEN_TIMER);
+			task_wakeup(task, TASK_WOKEN_TIMER);
 		}
 		else if (task->expire != eb->key) {
 			/* task is not expired but its key doesn't match so let's
@@ -355,13 +375,11 @@ void wake_expired_tasks()
 			if (tick_isset(task->expire))
 				__task_queue(task, &timers);
 			HA_RWLOCK_WRTOSK(TASK_WQ_LOCK, &wq_lock);
-			task_drop_running(task, 0);
 			goto lookup_next;
 		}
 		else {
 			/* task not expired and correctly placed. It may not be eternal. */
 			BUG_ON(task->expire == TICK_ETERNITY);
-			task_drop_running(task, 0);
 			break;
 		}
 	}
@@ -423,10 +441,10 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 	struct list *tl_queues = sched->tasklets;
 	struct task *t;
 	uint8_t budget_mask = (1 << TL_CLASSES) - 1;
+	struct sched_activity *profile_entry = NULL;
 	unsigned int done = 0;
 	unsigned int queue;
 	unsigned int state;
-	char heavy_calls = 0;
 	void *ctx;
 
 	for (queue = 0; queue < TL_CLASSES;) {
@@ -452,7 +470,7 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 
 			if (unlikely(queue > TL_NORMAL &&
 				     budget_mask & (1 << TL_NORMAL) &&
-				     ((sched->rqueue_size > 0) ||
+				     (!eb_is_empty(&sched->rqueue) ||
 				      (global_tasks_mask & tid_bit)))) {
 				/* a task was woken up by a bulk tasklet or another thread */
 				break;
@@ -472,21 +490,6 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 		}
 
 		budgets[queue]--;
-
-		t = (struct task *)LIST_ELEM(tl_queues[queue].n, struct tasklet *, list);
-		if (t->state & TASK_HEAVY) {
-			/* This is a heavy task. We'll call no more than one
-			 * per function call. If we called one already, we'll
-			 * return and announce the max possible weight so that
-			 * the caller doesn't come back too soon.
-			 */
-			if (heavy_calls) {
-				done = INT_MAX;  // 11ms instead of 3 without this
-				break; // too many heavy tasks processed already
-			}
-			heavy_calls = 1;
-		}
-
 		ti->flags &= ~TI_FL_STUCK; // this thread is still running
 		activity[tid].ctxsw++;
 
@@ -496,15 +499,35 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 		t->calls++;
 		sched->current = t;
 
-		_HA_ATOMIC_SUB(&sched->rq_total, 1);
+		_HA_ATOMIC_DEC(&sched->rq_total);
+
+		if (t->state & TASK_F_TASKLET) {
+			uint64_t before = 0;
 
-		if (TASK_IS_TASKLET(t)) {
 			LIST_DEL_INIT(&((struct tasklet *)t)->list);
 			__ha_barrier_store();
 
+			if (unlikely(task_profiling_mask & tid_bit)) {
+				profile_entry = sched_activity_entry(sched_activity, t->process);
+				before = now_mono_time();
+#ifdef DEBUG_TASK
+				if (((struct tasklet *)t)->call_date) {
+					HA_ATOMIC_ADD(&profile_entry->lat_time, before - ((struct tasklet *)t)->call_date);
+					((struct tasklet *)t)->call_date = 0;
+				}
+#endif
+			}
+
 			state = _HA_ATOMIC_FETCH_AND(&t->state, TASK_PERSISTENT);
 			__ha_barrier_atomic_store();
+
 			process(t, ctx, state);
+
+			if (unlikely(task_profiling_mask & tid_bit)) {
+				HA_ATOMIC_INC(&profile_entry->calls);
+				HA_ATOMIC_ADD(&profile_entry->cpu_time, now_mono_time() - before);
+			}
+
 			done++;
 			sched->current = NULL;
 			__ha_barrier_store();
@@ -514,31 +537,24 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 		LIST_DEL_INIT(&((struct tasklet *)t)->list);
 		__ha_barrier_store();
 
-		/* We must be the exclusive owner of the TASK_RUNNING bit, and
-		 * have to be careful that the task is not being manipulated on
-		 * another thread finding it expired in wake_expired_tasks().
-		 * The TASK_RUNNING bit will be set during these operations,
-		 * they are extremely rare and do not last long so the best to
-		 * do here is to wait.
-		 */
-		state = _HA_ATOMIC_LOAD(&t->state);
-		do {
-			while (unlikely(state & TASK_RUNNING)) {
-				__ha_cpu_relax();
-				state = _HA_ATOMIC_LOAD(&t->state);
-			}
-		} while (!_HA_ATOMIC_CAS(&t->state, &state, (state & TASK_PERSISTENT) | TASK_RUNNING));
+		state = t->state;
+		while (!_HA_ATOMIC_CAS(&t->state, &state, (state & TASK_PERSISTENT) | TASK_RUNNING))
+			;
 
 		__ha_barrier_atomic_store();
 
 		/* OK then this is a regular task */
 
-		_HA_ATOMIC_SUB(&task_per_thread[tid].task_list_size, 1);
+		_HA_ATOMIC_DEC(&task_per_thread[tid].tasks_in_list);
 		if (unlikely(t->call_date)) {
 			uint64_t now_ns = now_mono_time();
+			uint64_t lat = now_ns - t->call_date;
 
-			t->lat_time += now_ns - t->call_date;
+			t->lat_time += lat;
 			t->call_date = now_ns;
+			profile_entry = sched_activity_entry(sched_activity, t->process);
+			HA_ATOMIC_ADD(&profile_entry->lat_time, lat);
+			HA_ATOMIC_INC(&profile_entry->calls);
 		}
 
 		__ha_barrier_store();
@@ -569,19 +585,22 @@ unsigned int run_tasks_from_lists(unsigned int budgets[])
 		 */
 		if (t != NULL) {
 			if (unlikely(t->call_date)) {
-				t->cpu_time += now_mono_time() - t->call_date;
+				uint64_t cpu = now_mono_time() - t->call_date;
+
+				t->cpu_time += cpu;
 				t->call_date = 0;
+				HA_ATOMIC_ADD(&profile_entry->cpu_time, cpu);
 			}
 
-			state = _HA_ATOMIC_LOAD(&t->state);
+			state = _HA_ATOMIC_AND_FETCH(&t->state, ~TASK_RUNNING);
 			if (unlikely(state & TASK_KILLED)) {
 				task_unlink_wq(t);
 				__task_free(t);
 			}
-			else {
+			else if (state & TASK_WOKEN_ANY)
+				task_wakeup(t, 0);
+			else
 				task_queue(t);
-				task_drop_running(t, 0);
-			}
 		}
 		done++;
 	}
@@ -613,12 +632,16 @@ void process_runnable_tasks()
 		[TL_URGENT] = 64, // ~50% of CPU bandwidth for I/O
 		[TL_NORMAL] = 48, // ~37% of CPU bandwidth for tasks
 		[TL_BULK]   = 16, // ~13% of CPU bandwidth for self-wakers
+		[TL_HEAVY]  = 1,  // never more than 1 heavy task at once
 	};
 	unsigned int max[TL_CLASSES]; // max to be run per class
 	unsigned int max_total;       // sum of max above
 	struct mt_list *tmp_list;
 	unsigned int queue;
 	int max_processed;
+	int lpicked, gpicked;
+	int heavy_queued = 0;
+	int budget;
 
 	ti->flags &= ~TI_FL_STUCK; // this thread is still running
 
@@ -627,7 +650,6 @@ void process_runnable_tasks()
 		return;
 	}
 
-	nb_tasks_cur = nb_tasks;
 	max_processed = global.tune.runqueue_depth;
 
 	if (likely(niced_tasks))
@@ -650,37 +672,57 @@ void process_runnable_tasks()
 
 	/* normal tasklets list gets a default weight of ~37% */
 	if ((tt->tl_class_mask & (1 << TL_NORMAL)) ||
-	    (sched->rqueue_size > 0) || (global_tasks_mask & tid_bit))
+	    !eb_is_empty(&sched->rqueue) || (global_tasks_mask & tid_bit))
 		max[TL_NORMAL] = default_weights[TL_NORMAL];
 
 	/* bulk tasklets list gets a default weight of ~13% */
 	if ((tt->tl_class_mask & (1 << TL_BULK)))
 		max[TL_BULK] = default_weights[TL_BULK];
 
+	/* heavy tasks are processed only once and never refilled in a
+	 * call round. That budget is not lost either as we don't reset
+	 * it unless consumed.
+	 */
+	if (!heavy_queued) {
+		if ((tt->tl_class_mask & (1 << TL_HEAVY)))
+			max[TL_HEAVY] = default_weights[TL_HEAVY];
+		else
+			max[TL_HEAVY] = 0;
+		heavy_queued = 1;
+	}
+
 	/* Now compute a fair share of the weights. Total may slightly exceed
 	 * 100% due to rounding, this is not a problem. Note that while in
 	 * theory the sum cannot be NULL as we cannot get there without tasklets
 	 * to process, in practice it seldom happens when multiple writers
-	 * conflict and rollback on MT_LIST_TRY_ADDQ(shared_tasklet_list), causing
+	 * conflict and rollback on MT_LIST_TRY_APPEND(shared_tasklet_list), causing
 	 * a first MT_LIST_ISEMPTY() to succeed for thread_has_task() and the
 	 * one above to finally fail. This is extremely rare and not a problem.
 	 */
-	max_total = max[TL_URGENT] + max[TL_NORMAL] + max[TL_BULK];
+	max_total = max[TL_URGENT] + max[TL_NORMAL] + max[TL_BULK] + max[TL_HEAVY];
 	if (!max_total)
 		return;
 
 	for (queue = 0; queue < TL_CLASSES; queue++)
 		max[queue]  = ((unsigned)max_processed * max[queue] + max_total - 1) / max_total;
 
+	/* The heavy queue must never process more than one task at once
+	 * anyway.
+	 */
+	if (max[TL_HEAVY] > 1)
+		max[TL_HEAVY] = 1;
+
 	lrq = grq = NULL;
 
 	/* pick up to max[TL_NORMAL] regular tasks from prio-ordered run queues */
 	/* Note: the grq lock is always held when grq is not null */
-	while (tt->task_list_size < max[TL_NORMAL]) {
+	lpicked = gpicked = 0;
+	budget = max[TL_NORMAL] - tt->tasks_in_list;
+	while (lpicked + gpicked < budget) {
 		if ((global_tasks_mask & tid_bit) && !grq) {
 #ifdef USE_THREAD
 			HA_SPIN_LOCK(TASK_RQ_LOCK, &rq_lock);
-			grq = eb32sc_lookup_ge(&rqueue, rqueue_ticks - TIMER_LOOK_BACK, tid_bit);
+			grq = eb32sc_lookup_ge(&rqueue, global_rqueue_ticks - TIMER_LOOK_BACK, tid_bit);
 			if (unlikely(!grq)) {
 				grq = eb32sc_first(&rqueue, tid_bit);
 				if (!grq) {
@@ -696,7 +738,7 @@ void process_runnable_tasks()
 		 */
 
 		if (!lrq) {
-			lrq = eb32sc_lookup_ge(&tt->rqueue, rqueue_ticks - TIMER_LOOK_BACK, tid_bit);
+			lrq = eb32sc_lookup_ge(&tt->rqueue, tt->rqueue_ticks - TIMER_LOOK_BACK, tid_bit);
 			if (unlikely(!lrq))
 				lrq = eb32sc_first(&tt->rqueue, tid_bit);
 		}
@@ -707,13 +749,16 @@ void process_runnable_tasks()
 		if (likely(!grq || (lrq && (int)(lrq->key - grq->key) <= 0))) {
 			t = eb32sc_entry(lrq, struct task, rq);
 			lrq = eb32sc_next(lrq, tid_bit);
-			__task_unlink_rq(t);
+			eb32sc_delete(&t->rq);
+			lpicked++;
 		}
 #ifdef USE_THREAD
 		else {
 			t = eb32sc_entry(grq, struct task, rq);
 			grq = eb32sc_next(grq, tid_bit);
-			__task_unlink_rq(t);
+			_HA_ATOMIC_AND(&t->state, ~TASK_GLOBAL);
+			eb32sc_delete(&t->rq);
+
 			if (unlikely(!grq)) {
 				grq = eb32sc_first(&rqueue, tid_bit);
 				if (!grq) {
@@ -721,16 +766,14 @@ void process_runnable_tasks()
 					HA_SPIN_UNLOCK(TASK_RQ_LOCK, &rq_lock);
 				}
 			}
+			gpicked++;
 		}
 #endif
+		if (t->nice)
+			_HA_ATOMIC_DEC(&niced_tasks);
 
-		/* Make sure the entry doesn't appear to be in a list */
-		LIST_INIT(&((struct tasklet *)t)->list);
-		/* And add it to the local task list */
-		tasklet_insert_into_tasklet_list(&tt->tasklets[TL_NORMAL], (struct tasklet *)t);
-		tt->tl_class_mask |= 1 << TL_NORMAL;
-		_HA_ATOMIC_ADD(&tt->task_list_size, 1);
-		activity[tid].tasksw++;
+		/* Add it to the local task list */
+		LIST_APPEND(&tt->tasklets[TL_NORMAL], &((struct tasklet *)t)->list);
 	}
 
 	/* release the rqueue lock */
@@ -739,6 +782,18 @@ void process_runnable_tasks()
 		grq = NULL;
 	}
 
+	if (lpicked + gpicked) {
+		tt->tl_class_mask |= 1 << TL_NORMAL;
+		_HA_ATOMIC_ADD(&tt->tasks_in_list, lpicked + gpicked);
+#ifdef USE_THREAD
+		if (gpicked) {
+			_HA_ATOMIC_SUB(&grq_total, gpicked);
+			_HA_ATOMIC_ADD(&tt->rq_total, gpicked);
+		}
+#endif
+		activity[tid].tasksw += lpicked + gpicked;
+	}
+
 	/* Merge the list of tasklets waken up by other threads to the
 	 * main list.
 	 */
@@ -852,7 +907,7 @@ void mworker_cleantasks()
 /* perform minimal intializations */
 static void init_task()
 {
-	int i;
+	int i, q;
 
 #ifdef USE_THREAD
 	memset(&timers, 0, sizeof(timers));
@@ -860,16 +915,15 @@ static void init_task()
 #endif
 	memset(&task_per_thread, 0, sizeof(task_per_thread));
 	for (i = 0; i < MAX_THREADS; i++) {
-		LIST_INIT(&task_per_thread[i].tasklets[TL_URGENT]);
-		LIST_INIT(&task_per_thread[i].tasklets[TL_NORMAL]);
-		LIST_INIT(&task_per_thread[i].tasklets[TL_BULK]);
+		for (q = 0; q < TL_CLASSES; q++)
+			LIST_INIT(&task_per_thread[i].tasklets[q]);
 		MT_LIST_INIT(&task_per_thread[i].shared_tasklet_list);
 	}
 }
 
 /* config parser for global "tune.sched.low-latency", accepts "on" or "off" */
 static int cfg_parse_tune_sched_low_latency(char **args, int section_type, struct proxy *curpx,
-                                      struct proxy *defpx, const char *file, int line,
+                                      const struct proxy *defpx, const char *file, int line,
                                       char **err)
 {
 	if (too_many_args(1, args, err, NULL))
